import java.util.Vector;
import java.util.Iterator;

/**
 * Model of a TaxiDispatch
 * 
 * @author J. David Fleig
 * @version July 2005
 */
public class TaxiDispatch
{
    protected TaxiGarage garage;
    protected Vector taxisInService;
    protected Vector taxisWaiting;
    protected int minutes; 

    /**
     * Constructor for objects of class TaxiDispatch
     */
    public TaxiDispatch()
    {
        garage = new TaxiGarage();
        taxisInService = new Vector();
        taxisWaiting = new Vector();
    }
    /**
     * Gets the garage the dispatch is using for its taxis
     * 
     * @return The garage the Dispatch is using for its taxis
     */
    public TaxiGarage getGarage()
    {
        return garage;
    }
    /**
     * Gets the Vector of taxis which have customers
     * 
     * @return The taxis which are in service right now
     */
    public Vector getTaxisInService()
    {
        return taxisInService;
    }
     /**
     * Gets the Vector of taxis which are waiting
     * 
     * @return The taxis which are waiting
     */
    public Vector getTaxisWaiting()
    {
        return taxisWaiting;
    }
     /**
     * Gets the current time in minutes
     * 
     * @return The current time in minutes
     */
    public int getTime()
    {
        return minutes;
    }
    /**
     * Advances the time of day 
     * 
     * @param minutes The number of minutes to advance the clock by
     */
    public void advanceTime(int minutes)
    {
        this.minutes += minutes;     
        adjustTaxis();
    }
    
 
    private void adjustTaxis()
    {
         //examine all taxis in service to account for change in time.
        Iterator taxiIterator = taxisInService.iterator();
        while ( taxiIterator.hasNext() )
        {
            TaxiCab t = (TaxiCab) taxiIterator.next();
            int d = t.getDailyTimeInOperation();
            int diff = minutes - d;
            //If the diff is smaller than a full minute, don't update at all.
            while ( diff >= 1 )
            {
                adjustTaxi(t); //adjust its position for a full minute having passed
                t.setDailyTimeInOperation( t.getDailyTimeInOperation() + 1 ); //adjust its last updated clock for that min
                diff -= 1;
            }
            //check to see if it is available for hire now
            if (t.isForHire())
            {
                taxiIterator.remove();
                taxisWaiting.add(t);
            }
        }
    }
    private void adjustTaxi(TaxiCab t)
    {
        if (t.isForHire())
            return;
            
            if ( t.hasPassenger() )
            {
                //Taxi has passenger, should move toward destiation if not already there
                if ( t.getCurrentLocationX()==t.getDestinationLocationX() && t.getCurrentLocationY()==t.getDestinationLocationY() )
                {
                    //already at destination, drop passenger off
                    t.setHasPassenger(false);
                    t.setForHire(true);
                }
                else
                {
                    //move one spot toward destination
                    if ( t.getCurrentLocationX() < t.getDestinationLocationX() )
                    {
                        t.setCurrentLocation( t.getCurrentLocationX() + 1 , t.getCurrentLocationY() );
                    }
                    else if ( t.getCurrentLocationX() > t.getDestinationLocationX() )
                    {
                        t.setCurrentLocation( t.getCurrentLocationX() - 1 , t.getCurrentLocationY() );
                    }
                    else if ( t.getCurrentLocationY() < t.getDestinationLocationY() )
                    {
                         t.setCurrentLocation( t.getCurrentLocationX(), t.getCurrentLocationY() + 1  );
                    }
                    else if ( t.getCurrentLocationY() > t.getDestinationLocationY() )
                    {
                        t.setCurrentLocation( t.getCurrentLocationX(), t.getCurrentLocationY() - 1  );
                    }
                }
            }
            else
            {
                //Taxi does not have passenger, should move toward pickup if not already there
                if ( t.getCurrentLocationX()==t.getPickupLocationX() && t.getCurrentLocationY()==t.getPickupLocationY() )
                {
                    //pickup passenger
                    t.setHasPassenger(true);
                }
                else
                {
                    //move one spot toward pickup
                    if ( t.getCurrentLocationX() < t.getPickupLocationX() )
                    {
                        t.setCurrentLocation( t.getCurrentLocationX() + 1 , t.getCurrentLocationY() );
                    }
                    else if ( t.getCurrentLocationX() > t.getPickupLocationX() )
                    {
                        t.setCurrentLocation( t.getCurrentLocationX() - 1 , t.getCurrentLocationY() );
                    }
                    else if ( t.getCurrentLocationY() < t.getPickupLocationY() )
                    {
                         t.setCurrentLocation( t.getCurrentLocationX(), t.getCurrentLocationY() + 1  );
                    }
                    else if ( t.getCurrentLocationY() > t.getPickupLocationY() )
                    {
                        t.setCurrentLocation( t.getCurrentLocationX(), t.getCurrentLocationY()  );
                    }
                }
            }
        
    }
    
    /**
     * Request a Taxi
     * 
     * @throws TaxiNotAvailableException When there is no taxi that can fullfill the request
     * 
     * @param pickupX The X location the passenger wishes to be picked up
     * @param pickupY The Y location teh passenger wishes to be picked up
     * @param destinationX the X location the passenger wishes to be dropped off
     * @param destinationY the Y location the passenger wishes to be dropped off
     * @return The TaxiCab that will fullfill the request.
     * 
     */
    public TaxiCab requestTaxi(int pickupX, int pickupY, int destinationX, int destinationY)
     throws TaxiNotAvailableException
    {
        TaxiCab t=null;
        if (taxisWaiting.size() > 0)
        {
            //find the closest taxi in waiting and send it
            //right now we'll just send any taxi
            t = (TaxiCab) taxisWaiting.elementAt(0);
            taxisWaiting.remove(0);
            t.setPickupLocation(pickupX,pickupY);
            t.setDestinationLocation(destinationX,destinationY);
            t.setForHire(false);
            t.setHasPassenger(false);
            taxisInService.add(t);    
        }
        else if (garage.hasTaxiReady())
        {
            //get a taxi from the garage and send it
            t = garage.getTaxi();
            t.setPickupLocation(pickupX,pickupY);
            t.setDestinationLocation(destinationX,destinationY);
            t.setForHire(false);
            t.setHasPassenger(false);
            taxisInService.add(t);
        }
        else
        {
            throw new TaxiNotAvailableException();
        }
        
        t.setDailyTimeInOperation(minutes);
        return t;
    }
 
}
